@@ -315,9 +315,10 @@ class ANN:
         # return values and index of highest probability and random samples from pmf
         return probs, idx_max, None
 
-    def jacobian(self, X_i, batch_size=1, feed_forward=True):
+    def d_norm_y_dX(self, X_i, batch_size=1, feed_forward=True, norm = True):
         """
-        Compute the Jacobian of the neural net via back propagation.
+        Compute the derivatives of the squared L2 norm of the output wrt
+        the inputs.
 
         Parameters
         ----------
@@ -325,26 +326,32 @@ class ANN:
             The input features.
         batch_size : int, optional
             The batch size. The default is 1.
+        feed_forward: Boolean, optional, default is True.
+            Feed X_i forward before computing the gradient.
+        norm : Boolean, optional, default is True.
+            Compute the gradient of ||y||_2. If False it computes the gradient of
+            y, if y is a scalar. If False and y is a vector, the resulting gradient is the
+            column sum of the full Jacobian matrix.
 
         Returns
         -------
         array
-            The Jacobian.
+            The derivatives [d||y||_2/dX_1, ..., d||y||_2/dX_n_in]
 
         """
         if feed_forward:
             self.feed_forward(X_i, batch_size=batch_size)
 
         for i in range(self.n_layers, -1, -1):
-            self.layers[i].compute_delta_hy()
+            self.layers[i].compute_delta_hy(norm=norm)
             # also compute the gradient of the output wrt the weights
             if i > 0:
                 self.layers[i].compute_y_grad_W()
 
-        # delta_hy of the input layer = the Jacobian of the neural net
+        # delta_hy of the input layer = the derivative of the normed output
         return self.layers[0].delta_hy
 
-    def back_prop(self, y_i, jacobian=False):
+    def back_prop(self, y_i):
         """
         Back-propagation algorithm to find gradient of the loss function with respect
         to the weights of the neural network.
@@ -362,7 +369,7 @@ class ANN:
 
         # start back propagation over hidden layers, starting with output layer
         for i in range(self.n_layers, 0, -1):
-            self.layers[i].back_prop(y_i, jacobian=jacobian)
+            self.layers[i].back_prop(y_i)
 
     def batch(self, X_i, y_i, alpha=0.001, beta1=0.9, beta2=0.999):
         """
@@ -479,8 +486,8 @@ class ANN:
             if store_loss:
                 # l = 0.0
                 # for k in range(self.n_out):
-                    # l += self.layers[-1].L_i
-                
+                # l += self.layers[-1].L_i
+
                 l = self.layers[-1].L_i
 
                 if np.mod(i, 1000) == 0:
