@@ -10,4 +10,7 @@ from .kmn_surrogate import KMN_Surrogate
 from .reduced_surrogate import Reduced_Surrogate
 from .ann_surrogate import ANN_Surrogate
 from .DAS_network import DAS_network
-from .CumSum_Layer import CumSum_Layer
\ No newline at end of file
+from .CumSum_Layer import CumSum_Layer
+from .das_surrogate import DAS_Surrogate
+from .gp_surrogate import GP_Surrogate
+from .GP import GP
